@@ -3,6 +3,7 @@ extern crate slog;
 use anyhow::{anyhow, Result};
 
 mod commute;
+mod config;
 mod owned;
 mod stack;
 
@@ -18,19 +19,46 @@ pub struct Config<'a> {
     pub logger: &'a slog::Logger,
 }
 
-pub fn run(config: &Config) -> Result<()> {
+pub fn run(config: &mut Config) -> Result<()> {
     let repo = git2::Repository::open_from_env()?;
     debug!(config.logger, "repository found"; "path" => repo.path().to_str());
+
+    // here, we default to the git config value,
+    // if the flag was not provided in the CLI.
+    //
+    // in the future, we'd likely want to differentiate between
+    // a "non-provided" option, vs an explicit --no-<option>
+    // that disables a behavior, much like git does.
+    // e.g. user may want to overwrite a config value with
+    // --no-one-fixup-per-commit -- then, defaulting to the config value
+    // like we do here is no longer sufficient. but until then, this is fine.
+    //
+    config.one_fixup_per_commit |= config::one_fixup_per_commit(&repo);
+
     run_with_repo(config, &repo)
 }
 
 fn run_with_repo(config: &Config, repo: &git2::Repository) -> Result<()> {
-    let stack = stack::working_stack(&repo, config.base, config.force, config.logger)?;
+    let stack = stack::working_stack(repo, config.base, config.force, config.logger)?;
     if stack.is_empty() {
         crit!(config.logger, "No commits available to fix up, exiting");
         return Ok(());
     }
 
+    let autostage_enabled = config::auto_stage_if_nothing_staged(repo);
+    let index_was_empty = nothing_left_in_index(repo)?;
+    let mut we_added_everything_to_index = false;
+    if autostage_enabled && index_was_empty {
+        we_added_everything_to_index = true;
+
+        // no matter from what subdirectory we're executing,
+        // "." will still refer to the root workdir.
+        let pathspec = ["."];
+        let mut index = repo.index()?;
+        index.add_all(pathspec.iter(), git2::IndexAddOption::DEFAULT, None)?;
+        index.write()?;
+    }
+
     let mut diff_options = Some({
         let mut ret = git2::DiffOptions::new();
         ret.context_lines(0)
@@ -63,10 +91,7 @@ fn run_with_repo(config: &Config, repo: &git2::Repository) -> Result<()> {
         }
 
         let summary_counts = stack::summary_counts(&stack);
-        (
-            stack.into_iter().zip(diffs.into_iter()).collect(),
-            summary_counts,
-        )
+        (stack.into_iter().zip(diffs).collect(), summary_counts)
     };
 
     let mut head_tree = repo.head()?.peel_to_tree()?;
@@ -160,7 +185,7 @@ fn run_with_repo(config: &Config, repo: &git2::Repository) -> Result<()> {
             let mut commuted_old_path = old_path;
             let mut commuted_index_hunk = isolated_hunk;
 
-            'commit: for &(ref commit, ref diff) in &stack {
+            'commit: for (commit, diff) in &stack {
                 let c_logger = config.logger.new(o!(
                     "commit" => commit.id().to_string(),
                 ));
@@ -243,6 +268,8 @@ fn run_with_repo(config: &Config, repo: &git2::Repository) -> Result<()> {
         }
     }
 
+    let target_always_sha: bool = config::fixup_target_always_sha(repo);
+
     hunks_with_commit.sort_by_key(|h| h.dest_commit.id());
     // * apply all hunks that are going to be fixed up into `dest_commit`
     // * commit the fixup
@@ -256,7 +283,7 @@ fn run_with_repo(config: &Config, repo: &git2::Repository) -> Result<()> {
         .zip(hunks_with_commit.iter().skip(1).map(Some).chain([None]))
     {
         let new_head_tree = apply_hunk_to_tree(
-            &repo,
+            repo,
             &head_tree,
             &current.hunk_to_apply,
             &current.index_patch.old_path,
@@ -273,11 +300,14 @@ fn run_with_repo(config: &Config, repo: &git2::Repository) -> Result<()> {
             // https://docs.rs/git2/0.7.5/src/git2/repo.rs.html#998
             // https://libgit2.org/libgit2/#HEAD/group/commit/git_commit_create
             let dest_commit_id = current.dest_commit.id().to_string();
-            let dest_commit_locator = current
-                .dest_commit
-                .summary()
-                .filter(|&msg| summary_counts[msg] == 1)
-                .unwrap_or(&dest_commit_id);
+            let dest_commit_locator = match target_always_sha {
+                true => &dest_commit_id,
+                false => current
+                    .dest_commit
+                    .summary()
+                    .filter(|&msg| summary_counts[msg] == 1)
+                    .unwrap_or(&dest_commit_id),
+            };
             let diff = repo
                 .diff_tree_to_tree(Some(&head_commit.tree()?), Some(&new_head_tree), None)?
                 .stats()?;
@@ -307,11 +337,30 @@ fn run_with_repo(config: &Config, repo: &git2::Repository) -> Result<()> {
         }
     }
 
+    if autostage_enabled && we_added_everything_to_index {
+        // now that the fixup commits have been created,
+        // we should unstage the remaining changes from the index.
+
+        let mut index = repo.index()?;
+        index.read_tree(&head_tree)?;
+        index.write()?;
+    }
+
     if patches_considered == 0 {
-        warn!(
-            config.logger,
-            "No additions staged, try adding something to the index."
-        );
+        if index_was_empty && !we_added_everything_to_index {
+            warn!(
+                config.logger,
+                "No changes staged, try adding something \
+                 to the index or set {} = true",
+                config::AUTO_STAGE_IF_NOTHING_STAGED_CONFIG_NAME
+            );
+        } else {
+            warn!(
+                config.logger,
+                "Could not find a commit to fix up, use \
+                 --base to increase the search range."
+            )
+        }
     } else if config.and_rebase {
         use std::process::Command;
         // unwrap() is safe here, as we exit early if the stack is empty
@@ -321,7 +370,7 @@ fn run_with_repo(config: &Config, repo: &git2::Repository) -> Result<()> {
         assert!(number_of_parents <= 1);
 
         let mut command = Command::new("git");
-        command.args(&["rebase", "--interactive", "--autosquash"]);
+        command.args(["rebase", "--interactive", "--autosquash", "--autostash"]);
 
         if number_of_parents == 0 {
             command.arg("--root");
@@ -421,6 +470,19 @@ fn split_lines_after(content: &[u8], n: usize) -> (&[u8], &[u8]) {
     content.split_at(split_index)
 }
 
+fn nothing_left_in_index(repo: &git2::Repository) -> Result<bool> {
+    let stats = index_stats(repo)?;
+    let nothing = stats.files_changed() == 0 && stats.insertions() == 0 && stats.deletions() == 0;
+    Ok(nothing)
+}
+
+fn index_stats(repo: &git2::Repository) -> Result<git2::DiffStats> {
+    let head = repo.head()?.peel_to_tree()?;
+    let diff = repo.diff_tree_to_index(Some(&head), Some(&repo.index()?), None)?;
+    let stats = diff.stats()?;
+    Ok(stats)
+}
+
 #[cfg(test)]
 mod tests {
     use std::path::{Path, PathBuf};
@@ -503,18 +565,6 @@ lines
         ctx
     }
 
-    fn nothing_left_in_index(ctx: Context) {
-        let head = ctx.repo.head().unwrap().peel_to_tree().unwrap();
-        let diff = ctx
-            .repo
-            .diff_tree_to_index(Some(&head), Some(&ctx.repo.index().unwrap()), None)
-            .unwrap();
-        let stats = diff.stats().unwrap();
-        assert_eq!(stats.files_changed(), 0);
-        assert_eq!(stats.insertions(), 0);
-        assert_eq!(stats.deletions(), 0);
-    }
-
     #[test]
     fn multiple_fixups_per_commit() {
         let ctx = prepare_and_stage();
@@ -537,7 +587,7 @@ lines
         revwalk.push_head().unwrap();
         assert_eq!(revwalk.count(), 3);
 
-        nothing_left_in_index(ctx);
+        assert!(nothing_left_in_index(&ctx.repo).unwrap());
     }
 
     #[test]
@@ -562,6 +612,161 @@ lines
         revwalk.push_head().unwrap();
         assert_eq!(revwalk.count(), 2);
 
-        nothing_left_in_index(ctx);
+        assert!(nothing_left_in_index(&ctx.repo).unwrap());
+    }
+
+    fn autostage_common(ctx: &Context, file_path: &PathBuf) -> (PathBuf, PathBuf) {
+        // 1 modification w/o staging
+        let path = ctx.join(&file_path);
+        let contents = std::fs::read_to_string(&path).unwrap();
+        let modifications = format!("{contents}\nnew_line2");
+        std::fs::write(&path, &modifications).unwrap();
+
+        // 1 extra file
+        let fp2 = PathBuf::from("unrel.txt");
+        std::fs::write(ctx.join(&fp2), "foo").unwrap();
+
+        (path, fp2)
+    }
+
+    #[test]
+    fn autostage_if_index_was_empty() {
+        let (ctx, file_path) = prepare_repo();
+
+        // requires enabled config var
+        ctx.repo
+            .config()
+            .unwrap()
+            .set_bool(config::AUTO_STAGE_IF_NOTHING_STAGED_CONFIG_NAME, true)
+            .unwrap();
+
+        autostage_common(&ctx, &file_path);
+
+        // run 'git-absorb'
+        let drain = slog::Discard;
+        let logger = slog::Logger::root(drain, o!());
+        let config = Config {
+            dry_run: false,
+            force: false,
+            base: None,
+            and_rebase: false,
+            whole_file: false,
+            one_fixup_per_commit: false,
+            logger: &logger,
+        };
+        run_with_repo(&config, &ctx.repo).unwrap();
+
+        let mut revwalk = ctx.repo.revwalk().unwrap();
+        revwalk.push_head().unwrap();
+        assert_eq!(revwalk.count(), 2);
+
+        assert!(nothing_left_in_index(&ctx.repo).unwrap());
+    }
+
+    #[test]
+    fn do_not_autostage_if_index_was_not_empty() {
+        let (ctx, file_path) = prepare_repo();
+
+        // enable config var
+        ctx.repo
+            .config()
+            .unwrap()
+            .set_bool(config::AUTO_STAGE_IF_NOTHING_STAGED_CONFIG_NAME, true)
+            .unwrap();
+
+        let (_, fp2) = autostage_common(&ctx, &file_path);
+        // we stage the extra file - should stay in index
+        add(&ctx.repo, &fp2);
+
+        // run 'git-absorb'
+        let drain = slog::Discard;
+        let logger = slog::Logger::root(drain, o!());
+        let config = Config {
+            dry_run: false,
+            force: false,
+            base: None,
+            and_rebase: false,
+            whole_file: false,
+            one_fixup_per_commit: false,
+            logger: &logger,
+        };
+        run_with_repo(&config, &ctx.repo).unwrap();
+
+        let mut revwalk = ctx.repo.revwalk().unwrap();
+        revwalk.push_head().unwrap();
+        assert_eq!(revwalk.count(), 1);
+
+        assert_eq!(index_stats(&ctx.repo).unwrap().files_changed(), 1);
+    }
+
+    #[test]
+    fn do_not_autostage_if_not_enabled_by_config_var() {
+        let (ctx, file_path) = prepare_repo();
+
+        // disable config var
+        ctx.repo
+            .config()
+            .unwrap()
+            .set_bool(config::AUTO_STAGE_IF_NOTHING_STAGED_CONFIG_NAME, false)
+            .unwrap();
+
+        autostage_common(&ctx, &file_path);
+
+        // run 'git-absorb'
+        let drain = slog::Discard;
+        let logger = slog::Logger::root(drain, o!());
+        let config = Config {
+            dry_run: false,
+            force: false,
+            base: None,
+            and_rebase: false,
+            whole_file: false,
+            one_fixup_per_commit: false,
+            logger: &logger,
+        };
+        run_with_repo(&config, &ctx.repo).unwrap();
+
+        let mut revwalk = ctx.repo.revwalk().unwrap();
+        revwalk.push_head().unwrap();
+        assert_eq!(revwalk.count(), 1);
+
+        assert!(nothing_left_in_index(&ctx.repo).unwrap());
+    }
+
+    #[test]
+    fn fixup_message_always_commit_sha_if_configured() {
+        let ctx = prepare_and_stage();
+
+        ctx.repo
+            .config()
+            .unwrap()
+            .set_bool(config::FIXUP_TARGET_ALWAYS_SHA_CONFIG_NAME, true)
+            .unwrap();
+
+        // run 'git-absorb'
+        let drain = slog::Discard;
+        let logger = slog::Logger::root(drain, o!());
+        let config = Config {
+            dry_run: false,
+            force: false,
+            base: None,
+            and_rebase: false,
+            whole_file: false,
+            one_fixup_per_commit: true,
+            logger: &logger,
+        };
+        run_with_repo(&config, &ctx.repo).unwrap();
+        assert!(nothing_left_in_index(&ctx.repo).unwrap());
+
+        let mut revwalk = ctx.repo.revwalk().unwrap();
+        revwalk.push_head().unwrap();
+
+        let oids: Vec<git2::Oid> = revwalk.by_ref().collect::<Result<Vec<_>, _>>().unwrap();
+        assert_eq!(oids.len(), 2);
+
+        let commit = ctx.repo.find_commit(oids[0]).unwrap();
+        let actual_msg = commit.summary().unwrap();
+        let expected_msg = format!("fixup! {}", oids[1]);
+        assert_eq!(actual_msg, expected_msg);
     }
 }
