@@ -2,18 +2,7 @@ use anyhow::{anyhow, Result};
 
 use std::collections::HashMap;
 
-pub const MAX_STACK_CONFIG_NAME: &str = "absorb.maxStack";
-pub const MAX_STACK: usize = 10;
-
-fn max_stack(repo: &git2::Repository) -> usize {
-    match repo
-        .config()
-        .and_then(|config| config.get_i64(MAX_STACK_CONFIG_NAME))
-    {
-        Ok(max_stack) if max_stack > 0 => max_stack as usize,
-        _ => MAX_STACK,
-    }
-}
+use crate::config;
 
 pub fn working_stack<'repo>(
     repo: &'repo git2::Repository,
@@ -43,7 +32,9 @@ pub fn working_stack<'repo>(
 
     let base_commit = match user_provided_base {
         // https://github.com/rust-lang/rfcs/issues/1815
-        Some(commitish) => Some(repo.find_commit(repo.revparse_single(commitish)?.id())?),
+        // user_provided_base isn't guaranteed to be a commit hash, so peel until a
+        // commit is found.
+        Some(commitish) => Some(repo.revparse_single(commitish)?.peel_to_commit()?),
         None => None,
     };
 
@@ -87,7 +78,7 @@ pub fn working_stack<'repo>(
                 break;
             }
         }
-        if ret.len() == max_stack(repo) && user_provided_base.is_none() {
+        if ret.len() == config::max_stack(repo) && user_provided_base.is_none() {
             warn!(logger, "stack limit reached, use --base or configure absorb.maxStack to override";
                   "limit" => ret.len());
             break;
@@ -230,14 +221,17 @@ mod tests {
     #[test]
     fn test_stack_stops_at_configured_limit() {
         let (_dir, repo) = init_repo();
-        let commits = empty_commit_chain(&repo, "HEAD", &[], MAX_STACK + 2);
+        let commits = empty_commit_chain(&repo, "HEAD", &[], config::MAX_STACK + 2);
         repo.config()
             .unwrap()
-            .set_i64(MAX_STACK_CONFIG_NAME, (MAX_STACK + 1) as i64)
+            .set_i64(
+                config::MAX_STACK_CONFIG_NAME,
+                (config::MAX_STACK + 1) as i64,
+            )
             .unwrap();
 
         assert_stack_matches_chain(
-            MAX_STACK + 1,
+            config::MAX_STACK + 1,
             &working_stack(&repo, None, false, &empty_slog()).unwrap(),
             &commits,
         );
